/*
 * This file is part of qtutil
 * Copyright (c) 2000 David Deaven (deaven@execpc.com)
 *
 *   This program is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation; either version 2 of the License, or
 *   (at your option) any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program; if not, write to the Free Software
 *   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 * $Author: deaven $
 * $Id: QTAtom.cxx,v 1.3 2000/10/14 04:47:06 deaven Exp $
 */

#include <stdio.h>
#include <ctype.h>
#include <string.h>
#include <time.h>
#include "QTAtom.h"
#include "QTAtomImpl.h"

#define MAX_FILE_ATOMS 256

void QTAtomImpl::convert_int16(int16* data)
{
    byte* h = (byte*) data;
    int16 val = (h[0] << 8) | h[1];
    *data = val;
}

void QTAtomImpl::convert_int32(int32* data)
{
    byte* h = (byte*) data;
    int32 val = (h[0] << 24) | (h[1] << 16) | (h[2] << 8) | h[3];
    *data = val;
}

int QTAtomImpl::read_int32(FILE* f, int32* val)
{
    if (1 != fread(val, sizeof(int32), 1, f)) return -1;
    convert_int32(val);
    return 0;
}

void QTAtomImpl::convert_fixed32(fixed32* data)
{
    byte* h = (byte*) data;
    int integral = (h[0] << 8) | h[1];
    int fractional = (h[2] << 8) | h[3];
    double r = (double) fractional;
    r *= (1.0 / 65536.0);
    r += (double) integral;
    *data = (fixed32) r;
}

time_t QTAtomImpl::convert_mac_time(int32 mac_time)
{
    // mac_time is seconds from midnight, January 1, 1904
    time_t result = (time_t) mac_time;

#if 0
    struct tm tmac, tunix;
    bzero(&tmac, sizeof(tmac));
    tmac.tm_year = 4;
    tmac.tm_mday = 1;
    bzero(&tunix, sizeof(tunix));
    tunix.tm_year = 70;
    tunix.tm_mday = 1;
    double delta = difftime(timegm(&tunix), timegm(&tmac));
    printf("66 years is %f seconds?\n", delta);
#else
    long delta = 2082844800; // from above code
#endif
    result -= (time_t) delta;
    return result;
}

void QTAtomImpl::print_mac_time(int32 mac_time)
{
    time_t t = convert_mac_time(mac_time);
    char* s = ctime(&t);
    char* nl = strchr(s, '\n');
    if (nl) *nl = '\0';
    printf("%s", s);
}

unsigned long QTAtomImpl::read_buffer(void* start, void* end, unsigned long atom_offset)
{
    unsigned long len = ((byte*) end) - ((byte*) start);
    unsigned long datasize = Size - HeaderSize - atom_offset;
    if (len > datasize)
    {
	len = datasize;
    }
    fseek(filePtr(), Offset + HeaderSize + atom_offset, SEEK_SET);
    fread(start, len, 1, filePtr());
    return len;
}

QTAtomImpl** QTAtomImpl::children(int* count)
{
    *count = 0;

    if (HeaderSize < 8 || Size <= HeaderSize)
    {
	return NULL;
    }
    
    unsigned long cdata_offset = HeaderSize;
    unsigned long cdata_max_size = Size - HeaderSize;
    QTAtomImpl* impl[MAX_FILE_ATOMS];

    while ((impl[*count] = read_atom_at(Offset + cdata_offset, cdata_max_size)))
    {
	cdata_offset += impl[*count]->Size;
	cdata_max_size -= impl[*count]->Size;
	(*count)++;
    }

    QTAtomImpl** result = new QTAtomImpl*[*count];
    int i;
    for (i = 0; i < (*count); i++)
    {
	result[i] = impl[i];
    }
    return result;
}

void QTAtomImpl::print()
{
    printf("type \"%c%c%c%c\" size %ldb",
	Type[0], Type[1], Type[2], Type[3],
	(long) Size);
}



// ================================================================
class QTRootAtomImpl : public QTAtomImpl {
public:
    QTRootAtomImpl(FILE* fptr) :
	QTAtomImpl(fptr)
    {
	fseek(fptr, 0, SEEK_END);
	Size = ftell(fptr);
    }
    QTAtomImpl** children(int* count);
};

QTAtomImpl** QTRootAtomImpl::children(int* count)
{
    // in a file, read as many atoms as there are serially
    QTAtomImpl* impl[MAX_FILE_ATOMS];
    *count = 0;

    unsigned long offset = 0;
    unsigned long max_size = Size;
    while ((impl[*count] = read_atom_at(offset, max_size)))
    {
	offset += impl[*count]->Size;
	max_size -= impl[*count]->Size;
	(*count)++;
    }

    QTAtomImpl** result = new QTAtomImpl*[*count];
    int i;
    for (i = 0; i < (*count); i++)
    {
	result[i] = impl[i];
    }
    return result;
}



// ================================================================
const char* QTAtom::type() const
{
    return Impl ? Impl->Type : NULL;
}

unsigned long QTAtom::size() const
{
    return Impl ? Impl->Size : 0;
}

QTAtom* QTAtom::children(int* count)
{
    if (!Impl)
    {
	*count = 0;
	return NULL;
    }

    QTAtomImpl** impl_children = Impl->children(count);
    QTAtom* result = (*count) > 0 ? new QTAtom[(*count)] : NULL;
    
    int i;
    for (i = 0; i < (*count); i++)
    {
	result[i].Impl = impl_children[i];
    }

    delete[] impl_children;
    return result;
}

void QTAtom::print(int indent)
{
    int spaces = 2 * indent;
    char* buf = new char[spaces + 1];
    memset(buf, ' ', spaces);
    buf[spaces] = '\0';
    printf("%s", buf);
    if (Impl) Impl->print();
    printf("\n");
    int count;
    QTAtom* kids = children(&count);
    int i;
    for (i = 0; i < count; i++)
    {
	kids[i].print(indent + 1);
    }
    delete[] kids;
    delete[] buf;
}

QTAtom QTAtom::parseFile(FILE* fptr)
{
    QTAtomImpl* impl = new QTRootAtomImpl(fptr);
    return QTAtom(impl);
}

QTAtomImpl* QTAtomImpl::read_atom_at(unsigned long offset, unsigned long max_size)
{
    if (max_size == 0) return NULL;

    char type[4];
    int32 len;
    int32 header_size = 0;
    fseek(filePtr(), offset, SEEK_SET);

    if (read_int32(filePtr(), &len) < 0) return NULL;
    if (len < 8 || (max_size > 0 && len > max_size)) return NULL;
    header_size += 4;
    if (1 != fread(&type, sizeof(type), 1, filePtr())) return NULL;
    header_size += sizeof(type);
    if (len == 1)
    {
	if (read_int32(filePtr(), &len) < 0) return NULL;
	header_size += 4;
	if (len < 12 || (max_size > 0 && len > max_size)) return NULL;
	printf("extended size: %db\n", len);
    }
    else if (len == 0)
    {
	long cur = ftell(filePtr());
	if (0 != fseek(filePtr(), 0, SEEK_END)) return NULL;
	len = ftell(filePtr());
	fseek(filePtr(), cur, SEEK_SET);
	if (max_size > 0 && len > max_size) return NULL;
    }

    // reject non-printable types

    if (!isprint(type[0]) ||
	!isprint(type[1]) ||
	!isprint(type[2]) ||
	!isprint(type[3]))
    {
	return NULL;
    }

    // handle atoms as we know how

    QTAtomImpl* impl = NULL;

#define KNOWN_TYPE(TYPE) \
if (impl == NULL && !strncmp(#TYPE, type, 4)) { \
    impl = new QT ## TYPE (this, type, len, header_size, offset); }

    KNOWN_TYPE(mvhd)
    KNOWN_TYPE(tkhd)
    KNOWN_TYPE(hdlr)
    KNOWN_TYPE(dref)
    KNOWN_TYPE(stsd)
    KNOWN_TYPE(stsc)
    KNOWN_TYPE(stsz)
    KNOWN_TYPE(stco)

    if (impl == NULL)
    {
	// opaque atom
	impl = new QTAtomImpl(this, type, len, header_size, offset);
    }

    // skip this atom's data
    fseek(filePtr(), offset + len, SEEK_SET);

    return impl;
}

int QTAtom::findChildrenOfTypeRecurse(
    const char* t, QTAtom* atoms, int *rcount)
{
    if (QTAtomImpl::cmp32(type(), t))
    {
	atoms[(*rcount)++] = *this;
    }

    int count;
    QTAtom* kids = children(&count);

    int i;
    for (i = 0; i < count; i++)
    {
	kids[i].findChildrenOfTypeRecurse(t, atoms, rcount);
    }
    delete[] kids;

    return 0;
}

QTAtom* QTAtom::findChildrenOfType(const char* t, int* rcount)
{
    QTAtom* atoms = new QTAtom[256];
    *rcount = 0;

    findChildrenOfTypeRecurse(t, atoms, rcount);

    return atoms;
}

QTAtom QTAtom::findChildOfType(const char* t)
{
    QTAtom result;
    int count;
    QTAtom* matches = findChildrenOfType(t, &count);
    if (count > 0)
    {
	result = matches[0];
    }
    delete[] matches;
    return result;
}
